package java.time;

import java.util.Collections;
import java.util.Arrays;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.math.RoundingMode;
import java.math.BigDecimal;
import java.util.List;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ChronoField;
import java.time.temporal.ChronoUnit;
import java.time.temporal.Temporal;
import java.util.regex.Matcher;
import java.time.format.DateTimeParseException;
import java.util.Iterator;
import java.util.Objects;
import java.time.temporal.TemporalUnit;
import java.util.regex.Pattern;
import java.math.BigInteger;
import java.io.Serializable;
import java.time.temporal.TemporalAmount;

public final class Duration implements TemporalAmount, Comparable<Duration>, Serializable
{
    public static final Duration ZERO;
    private static final long serialVersionUID = 3078945930695997490L;
    private static final BigInteger BI_NANOS_PER_SECOND;
    private static final Pattern PATTERN;
    private final long seconds;
    private final int nanos;
    
    public static Duration ofDays(long x) {
        return create(Math.multiplyExact(x, 86400L), 0);
    }
    
    public static Duration ofHours(long x) {
        return create(Math.multiplyExact(x, 3600L), 0);
    }
    
    public static Duration ofMinutes(long x) {
        return create(Math.multiplyExact(x, 60L), 0);
    }
    
    public static Duration ofSeconds(long n) {
        return create(n, 0);
    }
    
    public static Duration ofSeconds(long x, long n) {
        return create(Math.addExact(x, Math.floorDiv(n, 1000000000L)), (int)Math.floorMod(n, 1000000000L));
    }
    
    public static Duration ofMillis(long n) {
        long n2 = n / 1000L;
        int n3 = (int)(n % 1000L);
        if (n3 < 0) {
            n3 += 1000;
            --n2;
        }
        return create(n2, n3 * 1000000);
    }
    
    public static Duration ofNanos(long n) {
        long n2 = n / 1000000000L;
        int n3 = (int)(n % 1000000000L);
        if (n3 < 0) {
            n3 = (int)((long)n3 + 1000000000L);
            --n2;
        }
        return create(n2, n3);
    }
    
    public static Duration of(long n, TemporalUnit temporalUnit) {
        return Duration.ZERO.plus(n, temporalUnit);
    }
    
    public static Duration from(TemporalAmount obj) {
        Objects.requireNonNull(obj, "amount");
        Duration duration = Duration.ZERO;
        for (TemporalUnit temporalUnit : obj.getUnits()) {
            duration = duration.plus(obj.get(temporalUnit), temporalUnit);
        }
        return duration;
    }
    
    public static Duration parse(CharSequence charSequence) {
        Objects.requireNonNull(charSequence, "text");
        Matcher matcher = Duration.PATTERN.matcher(charSequence);
        if (matcher.matches() && !"T".equals(matcher.group(3))) {
            int equals = "-".equals(matcher.group(1));
            String group = matcher.group(2);
            String group2 = matcher.group(4);
            String group3 = matcher.group(5);
            String group4 = matcher.group(6);
            String group5 = matcher.group(7);
            if (group != null || group2 != null || group3 != null || group4 != null) {
                long number = parseNumber(charSequence, group, 86400, "days");
                long number2 = parseNumber(charSequence, group2, 3600, "hours");
                long number3 = parseNumber(charSequence, group3, 60, "minutes");
                long number4 = parseNumber(charSequence, group4, 1, "seconds");
                int fraction = parseFraction(charSequence, group5, (number4 < 0L) ? -1 : 1);
                try {
                    return create((boolean)equals, number, number2, number3, number4, fraction);
                }
                catch (ArithmeticException cause) {
                    throw (DateTimeParseException)new DateTimeParseException("Text cannot be parsed to a Duration: overflow", charSequence, 0).initCause(cause);
                }
            }
        }
        throw new DateTimeParseException("Text cannot be parsed to a Duration", charSequence, 0);
    }
    
    private static long parseNumber(CharSequence parsedData, String s, int n, String str) {
        if (s == null) {
            return 0L;
        }
        try {
            return Math.multiplyExact(Long.parseLong(s), (long)n);
        }
        catch (NumberFormatException | ArithmeticException cause) {
            throw (DateTimeParseException)new DateTimeParseException("Text cannot be parsed to a Duration: " + str, parsedData, 0).initCause(cause);
        }
    }
    
    private static int parseFraction(CharSequence parsedData, String substring, int n) {
        if (substring == null || substring.length() == 0) {
            return 0;
        }
        try {
            substring = (substring + "000000000").substring(0, 9);
            return Integer.parseInt(substring) * n;
        }
        catch (NumberFormatException | ArithmeticException cause) {
            throw (DateTimeParseException)new DateTimeParseException("Text cannot be parsed to a Duration: fraction", parsedData, 0).initCause(cause);
        }
    }
    
    private static Duration create(boolean b, long x, long x2, long x3, long y, int n) {
        long addExact = Math.addExact(x, Math.addExact(x2, Math.addExact(x3, y)));
        if (b) {
            return ofSeconds(addExact, (long)n).negated();
        }
        return ofSeconds(addExact, (long)n);
    }
    
    public static Duration between(Temporal temporal, Temporal temporal2) {
        try {
            return ofNanos(temporal.until(temporal2, ChronoUnit.NANOS));
        }
        catch (DateTimeException | ArithmeticException ex) {
            long until = temporal.until(temporal2, ChronoUnit.SECONDS);
            long n;
            try {
                n = temporal2.getLong(ChronoField.NANO_OF_SECOND) - temporal.getLong(ChronoField.NANO_OF_SECOND);
                if (until > 0L && n < 0L) {
                    ++until;
                }
                else if (until < 0L && n > 0L) {
                    --until;
                }
            }
            catch (DateTimeException ex2) {
                n = 0L;
            }
            return ofSeconds(until, n);
        }
    }
    
    private static Duration create(long n, int n2) {
        if ((n | (long)n2) == 0x0L) {
            return Duration.ZERO;
        }
        return new Duration(n, n2);
    }
    
    private Duration(long seconds, int nanos) {
        this.seconds = seconds;
        this.nanos = nanos;
    }
    
    @Override
    public long get(TemporalUnit obj) {
        if (obj == ChronoUnit.SECONDS) {
            return this.seconds;
        }
        if (obj == ChronoUnit.NANOS) {
            return (long)this.nanos;
        }
        throw new UnsupportedTemporalTypeException("Unsupported unit: " + obj);
    }
    
    @Override
    public List<TemporalUnit> getUnits() {
        return Duration.DurationUnits.UNITS;
    }
    
    public boolean isZero() {
        return (this.seconds | (long)this.nanos) == 0x0L;
    }
    
    public boolean isNegative() {
        return this.seconds < 0L;
    }
    
    public long getSeconds() {
        return this.seconds;
    }
    
    public int getNano() {
        return this.nanos;
    }
    
    public Duration withSeconds(long n) {
        return create(n, this.nanos);
    }
    
    public Duration withNanos(int n) {
        ChronoField.NANO_OF_SECOND.checkValidIntValue((long)n);
        return create(this.seconds, n);
    }
    
    public Duration plus(Duration duration) {
        return this.plus(duration.getSeconds(), (long)duration.getNano());
    }
    
    public Duration plus(long n, TemporalUnit obj) {
        Objects.requireNonNull(obj, "unit");
        if (obj == ChronoUnit.DAYS) {
            return this.plus(Math.multiplyExact(n, 86400L), 0L);
        }
        if (obj.isDurationEstimated()) {
            throw new UnsupportedTemporalTypeException("Unit must not have an estimated duration");
        }
        if (n == 0L) {
            return this;
        }
        if ((obj instanceof ChronoUnit)) {
            switch ((ChronoUnit)obj) {
                case NANOS: {
                    return this.plusNanos(n);
                }
                case MICROS: {
                    return this.plusSeconds(n / 1000000000L * 1000L).plusNanos(n % 1000000000L * 1000L);
                }
                case MILLIS: {
                    return this.plusMillis(n);
                }
                case SECONDS: {
                    return this.plusSeconds(n);
                }
                default: {
                    return this.plusSeconds(Math.multiplyExact(obj.getDuration().seconds, n));
                }
            }
        }
        else {
            Duration multipliedBy = obj.getDuration().multipliedBy(n);
            return this.plusSeconds(multipliedBy.getSeconds()).plusNanos((long)multipliedBy.getNano());
        }
    }
    
    public Duration plusDays(long x) {
        return this.plus(Math.multiplyExact(x, 86400L), 0L);
    }
    
    public Duration plusHours(long x) {
        return this.plus(Math.multiplyExact(x, 3600L), 0L);
    }
    
    public Duration plusMinutes(long x) {
        return this.plus(Math.multiplyExact(x, 60L), 0L);
    }
    
    public Duration plusSeconds(long n) {
        return this.plus(n, 0L);
    }
    
    public Duration plusMillis(long n) {
        return this.plus(n / 1000L, n % 1000L * 1000000L);
    }
    
    public Duration plusNanos(long n) {
        return this.plus(0L, n);
    }
    
    private Duration plus(long y, long n) {
        if ((y | n) == 0x0L) {
            return this;
        }
        long addExact = Math.addExact(Math.addExact(this.seconds, y), n / 1000000000L);
        n %= 1000000000L;
        return ofSeconds(addExact, (long)this.nanos + n);
    }
    
    public Duration minus(Duration duration) {
        long seconds = duration.getSeconds();
        int nano = duration.getNano();
        if (seconds == Long.MIN_VALUE) {
            return this.plus(Long.MAX_VALUE, (long)(-nano)).plus(1L, 0L);
        }
        return this.plus(-seconds, (long)(-nano));
    }
    
    public Duration minus(long n, TemporalUnit temporalUnit) {
        return (n == Long.MIN_VALUE) ? this.plus(Long.MAX_VALUE, temporalUnit).plus(1L, temporalUnit) : this.plus(-n, temporalUnit);
    }
    
    public Duration minusDays(long n) {
        return (n == Long.MIN_VALUE) ? this.plusDays(Long.MAX_VALUE).plusDays(1L) : this.plusDays(-n);
    }
    
    public Duration minusHours(long n) {
        return (n == Long.MIN_VALUE) ? this.plusHours(Long.MAX_VALUE).plusHours(1L) : this.plusHours(-n);
    }
    
    public Duration minusMinutes(long n) {
        return (n == Long.MIN_VALUE) ? this.plusMinutes(Long.MAX_VALUE).plusMinutes(1L) : this.plusMinutes(-n);
    }
    
    public Duration minusSeconds(long n) {
        return (n == Long.MIN_VALUE) ? this.plusSeconds(Long.MAX_VALUE).plusSeconds(1L) : this.plusSeconds(-n);
    }
    
    public Duration minusMillis(long n) {
        return (n == Long.MIN_VALUE) ? this.plusMillis(Long.MAX_VALUE).plusMillis(1L) : this.plusMillis(-n);
    }
    
    public Duration minusNanos(long n) {
        return (n == Long.MIN_VALUE) ? this.plusNanos(Long.MAX_VALUE).plusNanos(1L) : this.plusNanos(-n);
    }
    
    public Duration multipliedBy(long val) {
        if (val == 0L) {
            return Duration.ZERO;
        }
        if (val == 1L) {
            return this;
        }
        return create(this.toSeconds().multiply(BigDecimal.valueOf(val)));
    }
    
    public Duration dividedBy(long val) {
        if (val == 0L) {
            throw new ArithmeticException("Cannot divide by zero");
        }
        if (val == 1L) {
            return this;
        }
        return create(this.toSeconds().divide(BigDecimal.valueOf(val), RoundingMode.DOWN));
    }
    
    private BigDecimal toSeconds() {
        return BigDecimal.valueOf(this.seconds).add(BigDecimal.valueOf((long)this.nanos, 9));
    }
    
    private static Duration create(BigDecimal bigDecimal) {
        BigInteger bigIntegerExact = bigDecimal.movePointRight(9).toBigIntegerExact();
        BigInteger[] divideAndRemainder = bigIntegerExact.divideAndRemainder(Duration.BI_NANOS_PER_SECOND);
        if (divideAndRemainder[0].bitLength() > 63) {
            throw new ArithmeticException("Exceeds capacity of Duration: " + bigIntegerExact);
        }
        return ofSeconds(divideAndRemainder[0].longValue(), (long)divideAndRemainder[1].intValue());
    }
    
    public Duration negated() {
        return this.multipliedBy(-1L);
    }
    
    public Duration abs() {
        return this.isNegative() ? this.negated() : this;
    }
    
    @Override
    public Temporal addTo(Temporal temporal) {
        if (this.seconds != 0L) {
            temporal = temporal.plus(this.seconds, ChronoUnit.SECONDS);
        }
        if (this.nanos != 0) {
            temporal = temporal.plus((long)this.nanos, ChronoUnit.NANOS);
            return temporal;
        }
        return temporal;
    }
    
    @Override
    public Temporal subtractFrom(Temporal temporal) {
        if (this.seconds != 0L) {
            temporal = temporal.minus(this.seconds, ChronoUnit.SECONDS);
        }
        if (this.nanos != 0) {
            temporal = temporal.minus((long)this.nanos, ChronoUnit.NANOS);
            return temporal;
        }
        return temporal;
    }
    
    public long toDays() {
        return this.seconds / 86400L;
    }
    
    public long toHours() {
        return this.seconds / 3600L;
    }
    
    public long toMinutes() {
        return this.seconds / 60L;
    }
    
    public long toMillis() {
        return Math.addExact(Math.multiplyExact(this.seconds, 1000L), (long)(this.nanos / 1000000));
    }
    
    public long toNanos() {
        return Math.addExact(Math.multiplyExact(this.seconds, 1000000000L), (long)this.nanos);
    }
    
    @Override
    public int compareTo(Duration duration) {
        int compare = Long.compare(this.seconds, duration.seconds);
        if (compare != 0) {
            return compare;
        }
        return this.nanos - duration.nanos;
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof Duration)) {
            Duration duration = (Duration)o;
            return this.seconds == duration.seconds && this.nanos == duration.nanos;
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return (int)(this.seconds ^ this.seconds >>> 32) + 51 * this.nanos;
    }
    
    @Override
    public String toString() {
        if (this == Duration.ZERO) {
            return "PT0S";
        }
        long lng = this.seconds / 3600L;
        int i = (int)(this.seconds % 3600L / 60L);
        int j = (int)(this.seconds % 60L);
        StringBuilder sb = new StringBuilder(24);
        sb.append("PT");
        if (lng != 0L) {
            sb.append(lng).append('H');
        }
        if (i != 0) {
            sb.append(i).append('M');
        }
        if (j == 0 && this.nanos == 0 && sb.length() > 2) {
            return sb.toString();
        }
        if (j < 0 && this.nanos > 0) {
            if (j == -1) {
                sb.append("-0");
            }
            else {
                sb.append(j + 1);
            }
        }
        else {
            sb.append(j);
        }
        if (this.nanos > 0) {
            int length = sb.length();
            if (j < 0) {
                sb.append(2000000000L - (long)this.nanos);
            }
            else {
                sb.append((long)this.nanos + 1000000000L);
            }
            while (sb.charAt(sb.length() - 1) == '0') {
                sb.setLength(sb.length() - 1);
            }
            sb.setCharAt(length, '.');
        }
        sb.append('S');
        return sb.toString();
    }
    
    private Object writeReplace() {
        return new Ser((byte)1, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        dataOutput.writeLong(this.seconds);
        dataOutput.writeInt(this.nanos);
    }
    
    static Duration readExternal(DataInput dataInput) throws IOException {
        return ofSeconds(dataInput.readLong(), (long)dataInput.readInt());
    }
    
    static {
        ZERO = new Duration(0L, 0);
        BI_NANOS_PER_SECOND = BigInteger.valueOf(1000000000L);
        PATTERN = Pattern.compile("([-+]?)P(?:([-+]?[0-9]+)D)?(T(?:([-+]?[0-9]+)H)?(?:([-+]?[0-9]+)M)?(?:([-+]?[0-9]+)(?:[.,]([0-9]{0,9}))?S)?)?", 2);
    }
    
    private static class DurationUnits
    {
        static final List<TemporalUnit> UNITS;
        
        static {
            UNITS = Collections.unmodifiableList(Arrays.asList(new TemporalUnit[] { ChronoUnit.SECONDS, ChronoUnit.NANOS }));
        }
    }
}
